/*
 * caps.cpp - Classes for Entity Capabilities (JEP-115)
 * Copyright (C) 2005  Remko Troncon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// TODO: 
//  - Implement Server Optimization support (Section 5).
//  - Implement consistency checking (Section 8).

#include <qstring.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qpair.h>

#include "caps.h"
#include "psiaccount.h"
#include "xmpp_tasks.h"
#include "im.h"


#define REQUEST_TIMEOUT 3000

using namespace XMPP;

// -----------------------------------------------------------------------------

/**
 * \class CapsManager::CapsSpec
 * \brief A class representing an entity capability specification.
 * An entity capability is a combination of a node, a version, and a set of
 * extensions.
 */

/**
 * Default constructor.
 */
CapsManager::CapsSpec::CapsSpec()
{
}


/**
 * \brief Basic constructor.
 * @param node the node
 * @param ven the version
 * @param ext the list of extensions (separated by spaces)
 */
CapsManager::CapsSpec::CapsSpec(const QString& node, const QString& ver, const QString& ext) : node_(node), ver_(ver), ext_(ext) 
{
}


/**
 * \brief Returns the node of the capabilities specification.
 */
const QString& CapsManager::CapsSpec::node() const 
{ 
	return node_; 
}


/**
 * \brief Returns the version of the capabilities specification.
 */
const QString& CapsManager::CapsSpec::version() const 
{ 
	return ver_; 
}


/**
 * \brief Returns the extensions of the capabilities specification.
 */
const QString& CapsManager::CapsSpec::extensions() const 
{ 
	return ext_; 
}


/**
 * \brief Flattens the caps specification into the set of 'simple' 
 * specifications.
 * A 'simple' specification is a specification with exactly one extension,
 * or with the version number as the extension.
 *
 * Example: A caps specification with node=http://psi-im.org, version=0.10,
 * and ext='achat vchat' would be expanded into the following list of specs:
 *	node=http://psi-im.org, ver=0.10, ext=0.10
 *	node=http://psi-im.org, ver=0.10, ext=achat
 *	node=http://psi-im.org, ver=0.10, ext=vchat
 */
CapsManager::CapsSpecs CapsManager::CapsSpec::flatten() const 
{
	CapsSpecs l;
	l.append(CapsSpec(node(),version(),version()));
	QStringList exts = QStringList::split(" ",extensions());
	for (QStringList::ConstIterator i = exts.begin(); i != exts.end(); ++i) {
		l.append(CapsSpec(node(),version(),*i));
	}
	return l;
}

bool CapsManager::CapsSpec::operator==(const CapsSpec& s) const 
{
	return (node() == s.node() && version() == s.version() && extensions() == s.extensions());
}

bool CapsManager::CapsSpec::operator!=(const CapsSpec& s) const 
{
	return !((*this) == s);
}

bool CapsManager::CapsSpec::operator<(const CapsSpec& s) const 
{
	return (node() != s.node() ? node() < s.node() :
			(version() != s.version() ? version() < s.version() : 
			 extensions() < s.extensions()));
}

// -----------------------------------------------------------------------------

CapsManager::CapsInfo::CapsInfo() : discovered_(false), pendingRequests_(0)
{
	updateLastSeen();
}

const QStringList& CapsManager::CapsInfo::features() const
{
	return features_;
}

const DiscoItem::Identities& CapsManager::CapsInfo::identities() const
{
	return identities_;
}

QStringList CapsManager::CapsInfo::jids() const
{
	QStringList jids;
	
	QValueList<QPair<QString,PsiAccount*> >::ConstIterator j = jids_.begin();
	for( ; j != jids_.end(); ++j) {
		QString jid((*j).first);
		if (!jids.contains(jid))
			jids.push_back(jid);
	}
	return jids;
}

bool CapsManager::CapsInfo::discovered() const
{
	return discovered_;
}

int CapsManager::CapsInfo::pendingRequests() const
{
	return pendingRequests_;
}

void CapsManager::CapsInfo::reset()
{
	features_.clear();
	identities_.clear();
	discovered_ = false;
}

void CapsManager::CapsInfo::removeAccount(PsiAccount* acc)
{
	QValueList<QPair<QString,PsiAccount*> >::Iterator j = jids_.begin();
	while (j != jids_.end()) {
		if ((*j).second == acc) {
			//qDebug(QString("caps.cpp: Removing pair ('%1','%2')").arg(QString((*j).first).replace('%',"%%")).arg(acc->name()));
			QValueList<QPair<QString,PsiAccount*> >::Iterator oj = j;
			j++;
			jids_.remove(oj);
		}
		else {
			j++;
		}
	}
}

void CapsManager::CapsInfo::addJid(const Jid& jid, PsiAccount* account)
{
	QPair<QString,PsiAccount*> p(jid.full(),account);
	if (!jids_.contains(p)) {
		//qDebug(QString("caps.cpp: Adding %1 %2").arg(QString(jid.full()).replace('%',"%%")).arg((int) account));
		jids_.push_back(p);
		updateLastSeen();
	}
}

void CapsManager::CapsInfo::removeJid(const Jid& jid)
{
	qDebug(QString("caps.cpp: Unregistering %1").arg(QString(jid.full()).replace('%',"%%")));
	QValueList<QPair<QString,PsiAccount*> >::Iterator j = jids_.begin();
	while (j != jids_.end()) {
		if ((*j).first == jid.full()) {
			QValueList<QPair<QString,PsiAccount*> >::Iterator oj = j;
			j++;
			jids_.remove(oj);
		}
		else {
			j++;
		}
	}
}

QPair<Jid,PsiAccount*> CapsManager::CapsInfo::nextJid(const Jid& jid, const Task* t)
{
	qDebug("caps.cpp: Looking for next JID");
	QValueList<QPair<QString,PsiAccount*> >::ConstIterator j = jids_.begin();
	for( ; j != jids_.end(); ++j) {
		if ((*j).first == jid.full() && (*j).second->client()->rootTask() == t) {
			j++;
			if (j == jids_.end()) {
				qDebug("caps.cpp: No more JIDs");
				return QPair<Jid,PsiAccount*>(Jid(),NULL);
			}
			else if ((*j).second->isActive()) {
				qDebug("caps.cpp: Account isn't active");
				return QPair<Jid,PsiAccount*>((*j).first,(*j).second);
			}
		}
	}
	return QPair<Jid,PsiAccount*>(Jid(),NULL);
}

void CapsManager::CapsInfo::setDiscovered(bool b)
{
	discovered_ = b;
}

void CapsManager::CapsInfo::setPendingRequests(int p)
{
	pendingRequests_ = p;
}

void CapsManager::CapsInfo::setIdentities(const DiscoItem::Identities& i)
{
	identities_ = i;
}

void CapsManager::CapsInfo::setFeatures(const QStringList& f)
{
	features_ = f;
}
	
void CapsManager::CapsInfo::updateLastSeen()
{
	lastSeen_ = QDate::currentDate();
}

QDomElement CapsManager::CapsInfo::toXml(QDomDocument *doc) const
{
	QDomElement info = doc->createElement("info");
	//info.setAttribute("last-seen",lastSeen_.toString(Qt::ISODate));

	// Identities
	for (DiscoItem::Identities::ConstIterator i = identities_.begin(); i != identities_.end(); ++i) {
		QDomElement identity = doc->createElement("identity");
		identity.setAttribute("category",(*i).category);
		identity.setAttribute("name",(*i).name);
		identity.setAttribute("type",(*i).type);
		info.appendChild(identity);
	}

	// Features
	for (QStringList::ConstIterator f = features_.begin(); f != features_.end(); f++) {
		QDomElement feature = doc->createElement("feature");
		feature.setAttribute("node",*f);
		info.appendChild(feature);
	}

	return info;
}

void CapsManager::CapsInfo::fromXml(const QDomElement& e)
{
	if (e.tagName() != "info") {
		qWarning("caps.cpp: Invalid info element");
		return;
	}
	
	//if (!e.attribute("last-seen").isEmpty())
	//	lastSeen_ = QDate::fromString(e.attribute("last-seen"),Qt::ISODate);

	for(QDomNode n = e.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement i = n.toElement();
		if(i.isNull()) {
			qWarning("caps.cpp: Null element");
			continue;
		}

		if(i.tagName() == "identity") {
			DiscoItem::Identity id;
			id.category = i.attribute("category");
			id.name = i.attribute("name");
			id.type = i.attribute("type");
			identities_ += id;
		}
		else if (i.tagName() == "feature") {
			features_ += i.attribute("node");
		}
		else {
			qWarning("caps.cpp: Unknown element");
		}

		discovered_ = true;
	}
}

// -----------------------------------------------------------------------------

/**
 * \class CapsManager
 * \brief A singleton class managing all the capabilities of JIDs and their
 * clients.
 */

/**
 * \brief Default constructor.
 */
CapsManager::CapsManager() : isEnabled_(true) /*, pending_(0)*/
{
	//connect(&timer_,SIGNAL(timeout()),SLOT(timer_timeout()));
}


/**
 * \brief Checks whether the caps manager is enabled (and does lookups).
 */
bool CapsManager::isEnabled()
{
	return isEnabled_;
}

/**
 * \brief Enables or disables the caps manager.
 */
void CapsManager::setEnabled(bool b)
{
	isEnabled_ = b;
}

/**
 * \brief Resets all the capabilities information.
 * Currently doesn't re-start the lookup.
 */
void CapsManager::reset()
{
	QValueList<CapsInfo> info = capsInfo_.values();
	for(QValueList<CapsInfo>::Iterator i = info.begin(); i != info.end(); ++i) {
		(*i).reset();
	}
}

/**
 * \brief Unregisters an account such that it isn't used for caps lookups.
 */
void CapsManager::removeAccount(PsiAccount* acc)
{
	qDebug(QString("Removing account '%1'").arg(acc->name()));
	QValueList<CapsInfo> info = capsInfo_.values();
	for(QValueList<CapsInfo>::Iterator i = info.begin(); i != info.end(); ++i) {
		(*i).removeAccount(acc);
	}
}

/**
 * \brief Registers new incoming capabilities information of a JID.
 * If the features of the entity are unknown, discovery requests are sent to 
 * retrieve the information.
 *
 * @param acc The account from which the new info is coming. This will be used
 *		to do disco requests if needed.
 * @param jid The entity's JID 
 * @param node The entity's caps node
 * @param ver The entity's caps version
 * @param ext The entity's caps extensions
 */
void CapsManager::updateCaps(PsiAccount* acc, const Jid& jid, const QString& node, const QString& ver, const QString& ext)
{
	if (jid.compare(acc->jid(),false))
		return;

	CapsSpec c(node,ver,ext);
	if (capsSpecs_[jid.full()] != c) {
		//qDebug(QString("caps.cpp: Updating caps for %1").arg(QString(jid.full()).replace('%',"%%")));
		
		// Unregister from all old caps nodes
		// FIXME: We should only unregister & register from changed nodes
		CapsSpecs old_caps = capsSpecs_[jid.full()].flatten();
		for (CapsSpecs::ConstIterator oc = old_caps.begin(); oc != old_caps.end(); ++oc) {
			if ((*oc) != CapsSpec()) {
				capsInfo_[*oc].removeJid(jid);
			}
		}

		if (!node.isEmpty() && !ver.isEmpty()) {
			// Register with all new caps nodes
			capsSpecs_[jid.full()] = c;
			CapsSpecs caps = c.flatten();
			for (CapsSpecs::ConstIterator nc = caps.begin(); nc != caps.end(); ++nc) {
				capsInfo_[*nc].addJid(jid,acc);
			}
			
			emit capsChanged(jid); 

			// Register new caps and check if we need to discover features
			if (isEnabled()) {
				for (CapsSpecs::ConstIterator i = caps.begin(); i != caps.end(); ++i) {
					if (!capsInfo_[*i].discovered() && capsInfo_[*i].pendingRequests() == 0) {
						qDebug(QString("caps.cpp: Sending disco request to %1, node=%2").arg(QString(jid.full()).replace('%',"%%")).arg(node + "#" + (*i).extensions()));
						capsInfo_[*i].setPendingRequests(1);
						requestDiscoInfo(acc,jid,node + "#" + (*i).extensions());
					}
					//else {
					//	printf("Using cached info for %s at node %s\n",jid.full().latin1(),QString(node + "#" + (*i).extensions()).latin1());
					//}
				}
			}
		}
		else {
			// Remove all caps specifications
			qWarning(QString("caps.cpp: Illegal caps info from %1: node=%2, ver=%3").arg(QString(jid.full()).replace('%',"%%")).arg(node).arg(ver));
			capsSpecs_.remove(jid.full());
		}
	}
	else {
		// Add to the list of jids
		CapsSpecs caps = c.flatten();
		for (CapsSpecs::ConstIterator nc = caps.begin(); nc != caps.end(); ++nc) {
			capsInfo_[*nc].addJid(jid,acc);
		}
	}
}

/**
 * \brief Removes all feature information for a given JID.
 *
 * @param jid The entity's JID 
 */
void CapsManager::disableCaps(const Jid& jid)
{
	qDebug(QString("caps.cpp: Disabling caps for %1.").arg(QString(jid.full()).replace('%',"%%")));
	if (capsEnabled(jid)) {
		CapsSpecs cs = capsSpecs_[jid.full()].flatten();
		for (CapsSpecs::ConstIterator i = cs.begin(); i != cs.end(); ++i) {
			capsInfo_[*i].removeJid(jid);
		}
		capsSpecs_.remove(jid.full());
		emit capsChanged(jid);
	}
}

/**
 * \brief Sends a disco#info request to a given node of a jid through an account.
 * When the request is finished, the discoFinished() slot is called.
 *
 * @param acc The account through which to send the disco request.
 * @param jid The target entity's JID 
 * @param node The target disco#info node
 */
void CapsManager::requestDiscoInfo(PsiAccount* acc, const Jid& jid, const QString& node) 
{
	JT_DiscoInfo* disco = new JT_DiscoInfo(acc->client()->rootTask());
	connect(disco, SIGNAL(finished()), SLOT(discoFinished()));
	disco->get(jid, node);
	//pending_++;
	//timer_.start(REQUEST_TIMEOUT,true);
	disco->go(true);
}


/**
 * \brief Sends a capsChanged() signal for all the jids for which new 
 * features have recently been discovered.
 */
//void CapsManager::updatePendingJIDs() 
//{
//	//qDebug(QString("caps.cpp: Updating %1 pending jids").arg(pending_jids_.count()));
//	while (!pending_jids_.isEmpty()) {
//		Jid jid(pending_jids_.first());
//		pending_jids_.pop_front();
//		emit capsChanged(jid);
//	}
//}


/**
 * \brief Called when a reply to disco#info request was received.
 * If the result was succesful, the resulting features are recorded in the
 * features database for the requested node, and all the affected jids are
 * put in the queue for update notification.
 * If the result was unsuccesful, another jid with the same capabilities is
 * selected and sent a disco#info query.
 */
void CapsManager::discoFinished()
{
	JT_DiscoInfo *disco = (JT_DiscoInfo*)sender();
	if (!disco)
		return;

	DiscoItem item = disco->item();
	Jid jid = disco->jid();
	qDebug(QString("caps.cpp: Disco response from %1, node=%2, success=%3").arg(QString(jid.full()).replace('%',"%%")).arg(disco->node()).arg(disco->success()));
	QStringList tokens = QStringList::split("#",disco->node());

	// Update features
	Q_ASSERT(tokens.count() == 2);
	QString node = tokens[0];
	QString ext = tokens[1];
	CapsSpec jid_cs = capsSpecs_[jid.full()];
	if (jid_cs.node() == node) {
		CapsSpec cs(node,jid_cs.version(),ext);

		if (disco->success()) {
			// Save identities & features
			capsInfo_[cs].setIdentities(item.identities());
			capsInfo_[cs].setFeatures(item.features().list());
			capsInfo_[cs].setPendingRequests(0);
			capsInfo_[cs].setDiscovered(true);

			// Save
			save();
			
			// Notify affected jids.
			QStringList jids = capsInfo_[cs].jids();
			for (QStringList::ConstIterator j = jids.begin() ; j != jids.end() ; j++) {
				//qDebug(QString("caps.cpp: Notifying %1.").arg(QString(*j).replace('%',"%%")));
				emit capsChanged(*j);
			}
		}
		else {
			QPair<Jid,PsiAccount*> p = capsInfo_[cs].nextJid(jid,disco->parent());
			if (p.second) {
				qDebug(QString("caps.cpp: Falling back on %1.").arg(QString(p.first.full()).replace('%',"%%")));
				requestDiscoInfo(p.second,p.first,disco->node());
			}
			else {
				qDebug("caps.cpp: No valid disco request avalable.");
				capsInfo_[cs].setPendingRequests(0);
			}
		}
	}
	else 
		qWarning(QString("caps.cpp: Current client node '%1' does not match response '%2'").arg(jid_cs.node()).arg(node));

	//for (unsigned int i = 0; i < item.features().list().count(); i++) 
	//	printf("    Feature: %s\n",item.features().list()[i].latin1());

	// Check pending requests
//	pending_ = (pending_ > 0 ? pending_-1 : 0);
//	if (!pending_) {
//		timer_.stop();
//		updatePendingJIDs();
//	}
}


/**
 * \brief Some disco#info requests were sent, but no reply was received in a 
 * certain period of time. We're giving up on those, and call 
 * updatePendingJIDs().
 */
//void CapsManager::timer_timeout() 
//{
	//qDebug(QString("caps.cpp: Timeout (%1 pending)").arg(pending_));
//	if (pending_ > 0) {
//		pending_ = 0;
//		updatePendingJIDs();
//	}
//}


/**
 * \brief Checks whether a given JID is broadcastingn its entity capabilities.
 */
bool CapsManager::capsEnabled(const Jid& jid) const
{
	return capsSpecs_.contains(jid.full());
}


/**
 * \brief Requests the list of features of a given JID.
 */
XMPP::Features CapsManager::features(const Jid& jid) const
{
	//qDebug(QString("caps.cpp: Retrieving features of %1").arg(jid.full()));
	QStringList f;
	if (capsEnabled(jid)) {
		CapsSpecs cs = capsSpecs_[jid.full()].flatten();
		for (CapsSpecs::ConstIterator i = cs.begin(); i != cs.end(); ++i) {
			//qDebug(QString("    %1").arg(features_[*i].join("\n")));
			f += capsInfo_[*i].features();
		}
	}
	return Features(f);
}
	
/**
 * \brief Returns the client name of a given jid.
 */
QString CapsManager::clientName(const Jid& jid) const
{
	if (capsEnabled(jid)) {
		CapsSpec cs = capsSpecs_[jid.full()];
		QString name = capsInfo_[CapsSpec(cs.node(),cs.version(),cs.version())].identities().first().name;
		
		// Try to be intelligent about the name
		/*if (name.isEmpty()) {
			name = cs.node();
			if (name.startsWith("http://"))
				name = name.right(name.length() - 7);
				
			if (name.startsWith("www."))
				name = name.right(name.length() - 4);
			
			int cut_pos = name.find(".");
			if (cut_pos != -1) {
				name = name.left(cut_pos);
			}
		}*/

		return name;
	}
	else {
		return QString();
	}
}

/**
 * \brief Returns the client version of a given jid.
 */
QString CapsManager::clientVersion(const Jid& jid) const
{
	return (capsEnabled(jid) ? capsSpecs_[jid.full()].version() : QString());
}


/**
 * \brief Convert all capabilities info to XML.
 */
void CapsManager::save() const
{
	if (fileName_.isEmpty())
		return;

	// Generate XML
	QDomDocument doc;
	QDomElement capabilities = doc.createElement("capabilities");
	doc.appendChild(capabilities);
	QMap<CapsSpec,CapsInfo>::ConstIterator i = capsInfo_.begin();
	for( ; i != capsInfo_.end(); i++) {
		QDomElement info = i.data().toXml(&doc);
		info.setAttribute("node",i.key().node());
		info.setAttribute("ver",i.key().version());
		info.setAttribute("ext",i.key().extensions());
		capabilities.appendChild(info);
	}

	// Save
	QFile f(fileName_);
	if(!f.open(IO_WriteOnly)) {
		qDebug("caps.cpp: Unable to open caps file");
		return;
	}
	QTextStream t;
	t.setDevice(&f);
	t.setEncoding(QTextStream::UnicodeUTF8);
	t << doc.toString();
	t.unsetDevice();
	f.close();
}


/**
 * \brief Sets the file to save the capabilities info to
 */
void CapsManager::setFile(const QString& fileName)
{
	fileName_ = fileName;
	
	// Load settings
	QDomDocument doc;
	QFile f(fileName_);
	if (!f.open(IO_ReadOnly))
		return;
	if (!doc.setContent(&f))
		return;
	f.close();

	QDomElement caps = doc.documentElement();
	if (caps.tagName() != "capabilities") {
		qWarning("caps.cpp: Invalid capabilities element");
		return;
	}
		
	for(QDomNode n = caps.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement i = n.toElement();
		if(i.isNull()) {
			qWarning("caps.cpp: Null element");
			continue;
		}

		if(i.tagName() == "info") {
			CapsInfo info;
			info.fromXml(i);
			CapsSpec spec(i.attribute("node"),i.attribute("ver"),i.attribute("ext"));
			capsInfo_[spec] = info;
			//qDebug(QString("Read %1 %2 %3").arg(spec.node()).arg(spec.version()).arg(spec.extensions()));
		}
		else {
			qWarning("caps.cpp: Unknown element");
		}
	}
}


/**
 * \brief Retrieves the instance of the CapsManager.
 * If no instance existed, a new one is created.
 */
CapsManager* CapsManager::instance()
{
	if (!instance_) {
		instance_ = new CapsManager();
	}
	return instance_;
}

/**
 * \brief The instance of the CapsManager.
 */
CapsManager* CapsManager::instance_ = NULL;

